#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import numpy as np
import random as rand
from torchvision.transforms import *
F = functional
from PIL import Image

from ..toolkit import byte2arr


__all__ = ('Comburant', 'Random',
           'NEAREST', 'LINEAR', 'CUBIC', 'HORIZONTAL', 'VERTICAL',
           'Resize', 'Crop', 'Flip', 'Rotate',
           'Brighten', 'Contrast', 'Saturate', 'Hue')


NEAREST = 0
LINEAR = 1
CUBIC = 2

PIL_INTERP = {NEAREST: Image.NEAREST, LINEAR: Image.BILINEAR, CUBIC: Image.BICUBIC}

HORIZONTAL = 10
VERTICAL = 11




class Comburant(object):
    def __init__(self, *args, is_encoded=False):
        self.comburant = Compose(list(args))
        self.is_encoded = is_encoded

    def __call__(self, img):
        if self.is_encoded:
            img = byte2arr(img, as_np=False)
        img = self.comburant(img)
        img = np.array(img)
        img = img.astype(np.float32) / 255
        return img



class Random(object):
    def __init__(self, p, comburant):
        self.p = p
        self.cbr = comburant

    def __call__(self, img):
        if rand.random() < self.p:
            return self.cbr(img)
        else:
            return img



class Resize(object):
    def __init__(self, size, interp=LINEAR):
        # size: (height, width)
        self.size = size
        self.interp = interp

    def __call__(self, img):
        return img.resize((self.size[1], self.size[0]), PIL_INTERP[self.interp])



class Crop(object):
    def __init__(self, size, padding=(0, 0, 0, 0), area_ratio=(1, 1), aspect_ratio=(1, 1), interp=LINEAR):
        # size: (height, width)
        # padding: (left, top, right, bottom)
        self.size = size
        self.padding = padding
        self.area_ratio = area_ratio
        self.aspect_ratio = aspect_ratio
        if area_ratio == aspect_ratio == (1,1):
            self.comburant = RandomCrop(size)
        else:
            self.comburant = RandomResizedCrop(size, area_ratio, aspect_ratio, PIL_INTERP[interp])

    def __call__(self, img):
        img = F.pad(img, self.padding, 0, 'constant')
        # pad the width if needed
        if img.size[0] < self.size[1]:
            img = F.pad(img, (self.size[1] - img.size[0], 0), 0, 'constant')
        # pad the height if needed
        if img.size[1] < self.size[0]:
            img = F.pad(img, (0, self.size[0] - img.size[1]), 0, 'constant')
        return self.comburant(img)



class Flip(object):
    def __init__(self, axial):
        if axial == HORIZONTAL:
            self.comburant = RandomVerticalFlip(1)
        elif axial == VERTICAL:
            self.comburant = RandomHorizontalFlip(1)
        else:
            raise Exception('NEBULAE ERROR ⨷ the invoked flip type is not defined or supported.')

    def __call__(self, img):
        return self.comburant(img)



class Rotate(object):
    def __init__(self, degree, intact=False, interp=NEAREST):
        '''
        Args
        intact: whether to keep image intact which might enlarge the output size
        '''
        self.comburant = RandomRotation(degree, PIL_INTERP[interp], intact)

    def __call__(self, img):
        return self.comburant(img)



class Brighten(object):
    def __init__(self, range):
        self.comburant = ColorJitter(brightness=range)

    def __call__(self, img):
        return self.comburant(img)



class Contrast(object):
    def __init__(self, range):
        self.comburant = ColorJitter(contrast=range)

    def __call__(self, img):
        return self.comburant(img)



class Saturate(object):
    def __init__(self, range):
        self.comburant = ColorJitter(saturation=range)

    def __call__(self, img):
        return self.comburant(img)



class Hue(object):
    def __init__(self, range):
        self.comburant = ColorJitter(hue=range)

    def __call__(self, img):
        return self.comburant(img)


#!/usr/bin/env python
'''
component_pt
Created by Seria at 05/02/2019 1:41 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========

'''
# -*- coding:utf-8 -*-
from math import ceil, sqrt
import torch
import torch.nn as nn
import torch.nn.functional as F
from ..cockpit.engine_pt import EnginePT

__all__ = ('Craft', 'Rudder', 'Nozzle',
           'NEAREST', 'LINEAR', 'CUBIC',
           'Void', 'XavierNorm', 'XavierUnif', 'Normal', 'Uniform', 'Orthog', 'Zeros', 'Ones',
           'Conv', 'TransConv', 'Dense', 'Embed', 'Identity',
           'Mean', 'Max', 'Min', 'Sum', 'MatMul', 'Grad',
           'Reshape', 'Permute', 'Upscale', 'MaxPool', 'AvgPool',
           'EMA',
           'Sqrt',
           'Concat',
           'Clip', 'Dropout', 'BN', 'CBN', 'IN', 'CIN', 'LN', 'SN',
           'Relu', 'LRelu', 'Tanh', 'Sigm', 'Sftm', 'Sftp',
           'MAE', 'MSE', 'SigmXE', 'SftmXE',
           'AccCls',
           'StepLR', 'PolyLR', 'CosLR', 'ExpLR', 'WavyLR',
           'Momentum', 'Nesterov', 'RMSProp', 'Adam', 'AdaBelief')



NEAREST = 0
LINEAR = 1
CUBIC = 2
PT_INTERP = {NEAREST: 'nearest', LINEAR: 'bilinear', CUBIC: 'bicubic'}




class Craft(nn.Module):
    def __init__(self, engine, scope):
        super(Craft, self).__init__()
        self.engine = engine
        self.scope = scope
        self.__pods = []
        self.__dict = {}
        self.__formulated = False

    def run(self, *args, **kwargs):
        raise NotImplementedError

    def forward(self, *args, **kwargs):
        # assert all([isinstance(a, Tensor) for a in args])
        return self.run(*args, **kwargs)

    def gear(self, gr=None):
        if isinstance(gr, bool):
            if gr:
                self.train()
            else:
                self.eval()
        elif gr is None:
            gr = self.engine
            if gr.param['device'] == 'gpu':
                if gr.param['rank'] < 0:
                    self.cuda()
                else:
                    self.to(gr.device[gr.param['rank']])
            elif gr.param['device'] == 'cpu':
                self.cpu()
        else:
            raise Exception('NEBULAE ERROR ⨷ %s is not a valid type of gear.' % type(gr))

    def vars(self):
        return self.parameters()

    @property
    def pods(self):
        return self.__pods

    @pods.setter
    def pods(self, pods):
        if not self.__formulated:
            self.__pods = pods
            self.__formulated = True

    def __getitem__(self, key):
        paths = key.split('/')
        craft = self
        for p in paths[:-1]:
            craft = getattr(craft, p)
        if paths[-1] == '':
            return craft
        else:
            return craft.__dict[paths[-1]]

    def __setitem__(self, key, value):
        self.__dict[key] = value



class Rudder(object):
    def __init__(self):
        self.grader = torch.enable_grad()

    def __enter__(self):
        self.grader.__enter__()
        return True

    def __exit__(self, *args):
        self.grader.__exit__()

class Nozzle(object):
    def __init__(self):
        self.grader = torch.no_grad()

    def __enter__(self):
        self.grader.__enter__()
        return False

    def __exit__(self, *args):
        self.grader.__exit__()



# -------------------------------------- Layer --------------------------------------- #

class Void(object):
    def __init__(self):
        pass

    def __call__(self, params):
        pass



class XavierNorm(object):
    def __init__(self):
        self.iniz = nn.init.xavier_normal_

    def __call__(self, params):
        self.iniz(params)



class XavierUnif(object):
    def __init__(self):
        self.iniz = nn.init.xavier_uniform_

    def __call__(self, params):
        self.iniz(params)



class Normal(object):
    def __init__(self, mean=0, std=1):
        self.mean = mean
        self.std = std
        self.iniz = nn.init.normal_

    def __call__(self, params):
        self.iniz(params, self.mean, self.std)



class Uniform(object):
    def __init__(self, min=-1, max=1):
        self.min_val = min
        self.max_val = max
        self.iniz = nn.init.uniform_

    def __call__(self, params):
        self.iniz(params, a=self.min_val, b=self.max_val)



class Orthog(object):
    def __init__(self):
        self.iniz = nn.init.orthogonal_

    def __call__(self, params):
        self.iniz(params)



class Ones(object):
    def __init__(self):
        self.iniz = nn.init.ones_

    def __call__(self, params):
        self.iniz(params)



class Zeros(object):
    def __init__(self):
        self.iniz = nn.init.zeros_

    def __call__(self, params):
        self.iniz(params)



class Conv(Craft):
    def __init__(self, in_chs, out_chs, kernel: tuple, stride=1, padding=0, dilation=1, group=1,
                 w_init=XavierNorm(), b_init=Zeros(), scope='CONV'):
        '''
        Args:
        - in_chs: input channel
        - out_chs: output channel
        - kernel: kernel size (must be a tuple)
        - stride: moving stride
        - padding: padding size
        - dilation: stride in atrous convolution
        - group: number of groups to be divided
        - w_init: weight initializer
        - w_param: options for initializing weight
        - b_init: bias initializer
        - b_param: options for initializing bias
        - scope: name scope
        '''
        super(Conv, self).__init__(None, scope)
        dim = len(kernel)
        if dim == 1:
            conv_fn = nn.Conv1d
            pad_fn = nn.ConstantPad1d
        elif dim == 2:
            conv_fn = nn.Conv2d
            pad_fn = nn.ConstantPad2d
        elif dim == 3:
            conv_fn = nn.Conv3d
            pad_fn = nn.ConstantPad3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)

        if isinstance(stride, int):
            stride = dim * [stride]
        if isinstance(dilation, int):
            dilation = dim * [dilation]
        if isinstance(padding, int):
            padding = 2*dim * [padding]

        self.pad = pad_fn(padding, 0)
        self.conv = conv_fn(in_chs, out_chs, kernel, stride=stride, dilation=dilation,
                            groups=group, bias=False if isinstance(b_init, Void) else True)
        w_init(self.conv.weight)
        b_init(self.conv.bias)

    def run(self, x):
        y = self.conv(self.pad(x))
        return y



class TransConv(Craft):
    def __init__(self, in_chs, out_chs, out_size, kernel: tuple, stride=1, padding=0, dilation=1, group=1,
                 w_init=XavierNorm(), b_init=Zeros(), scope='TRANSCONV'):
        '''
        Args:
        - in_chs: input channel
        - out_chs: output channel
        - out_size: output size
        - kernel: kernel size (must be a tuple)
        - stride: moving stride
        - padding: padding size
        - dilation: stride in atrous convolution
        - group: number of groups to be divided
        - w_init: weight initializer
        - w_param: options for initializing weight
        - b_init: bias initializer
        - b_param: options for initializing bias
        - scope: name scope
        '''
        super(TransConv, self).__init__(None, scope)
        dim = len(kernel)
        if dim == 1:
            conv_fn = nn.ConvTranspose1d
        elif dim == 2:
            conv_fn = nn.ConvTranspose2d
        elif dim == 3:
            conv_fn = nn.ConvTranspose3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d convolution is not supported.' % dim)

        if isinstance(stride, int):
            stride = dim * [stride]
        if isinstance(dilation, int):
            dilation = dim * [dilation]
        if isinstance(padding, int):
            padding = dim * [padding]
            # compensation = dim * [0]
        elif isinstance(padding, (list, tuple)):
            pad = padding
            padding = []
            # compensation = []
            for d in range(dim-1, -1, -1):
                n_elm = pad[2*d] + pad[2*d+1]
                half_elm = ceil(n_elm / 2)
                padding.append(half_elm)
                # compensation.append(n_elm%2)

        self.target_shape = (out_chs,) + out_size
        self.conv = conv_fn(in_chs, out_chs, kernel, stride=stride, padding=padding,
                            dilation=dilation, groups=group, bias=False if isinstance(b_init, Void) else True)
        w_init(self.conv.weight)
        b_init(self.conv.bias)

    def run(self, x):
        target_shape = (x.shape[0],) + self.target_shape
        y = self.conv(x, output_size=target_shape)
        return y



class Dense(Craft):
    def __init__(self, in_chs, out_chs, axis=-1,
                 w_init=XavierNorm(), b_init=Zeros(), scope='DENSE'):
        super(Dense, self).__init__(None, scope)
        if axis == 0:
            raise Exception('NEBULAE ERROR ⨷ you cannot apply dense layer along batch axis.')
        else:
            self.axis = axis

        self.fc = nn.Linear(in_chs, out_chs, bias=False if isinstance(b_init, Void) else True)
        w_init(self.fc.weight)
        b_init(self.fc.bias)

    def run(self, x):
        if self.axis == -1:
            y = self.fc(x)
        else:
            dim = x.ndim()
            permuted = [i for i in range(dim)]
            permuted = permuted[:self.axis] + permuted[self.axis + 1:] + [self.axis]
            x = x.transpose(*permuted)
            y = self.fc(x)
            permuted = [i for i in range(dim)]
            permuted = permuted[:self.axis] + [dim - 1] + permuted[self.axis:-1]
            y = y.transpose(*permuted)
        return y



class Embed(Craft):
    def __init__(self, ntoken, token_dim, scope='EMBED'):
        super(Embed, self).__init__(None, scope)
        self.embd = nn.Embedding(ntoken, token_dim)

    def run(self, x):
        y = self.embd(x)
        return y



class Identity(Craft):
    def __init__(self, scope='IDENTITY'):
        super(Identity, self).__init__(None, scope)

    def run(self, x):
        return x



# ---------------------------------- Manipulation ------------------------------------ #

class EMA(Craft):
    def __init__(self, hull, decay_fn=lambda x: 0.9, scope='EMA'):
        super(EMA, self).__init__(None, scope)
        self.counter = 0
        self.decay_fn = decay_fn
        self.hull = hull
        # self.var_keys = [k for k, _ in self.hull.named_parameters()]
        # self.buf_keys = [k for k, _ in self.hull.named_buffers()]

        # initialize shadow as hull itself
        self._backup()

    def __getattr__(self, name: str):
        if '_parameters' in self.__dict__:
            _parameters = self.__dict__['_parameters']
            if name in _parameters:
                return _parameters[name]
        if '_buffers' in self.__dict__:
            _buffers = self.__dict__['_buffers']
            if name in _buffers:
                return _buffers[name]
        if '_modules' in self.__dict__:
            modules = self.__dict__['_modules']
            if name in modules:
                return modules[name]
        if hasattr(self, 'hull'):
            if hasattr(self.hull, name):
                return getattr(self.hull, name)
        raise AttributeError("'{}' object has no attribute '{}'".format(type(self).__name__, name))

    def _backup(self):
        self.shadow = {}
        hull_params = self.hull.state_dict()
        for k, v in hull_params.items():
            self.shadow[k] = v.clone().detach()

    def update(self):
        self.counter += 1 # count calling times
        decay = self.decay_fn(self.counter)
        with torch.no_grad():
            hull_params = self.hull.state_dict()
            for key in self.shadow.keys():
                hull_params[key].data.copy_(hull_params[key].data * (1 - decay) + self.shadow[key].data * decay)
        self._backup()

    def run(self, *args, **kwargs):
        return self.hull(*args, **kwargs)



# ----------------------------------- Mathmatic -------------------------------------- #

class Sqrt(Craft):
    def __init__(self, scope='SQRT'):
        super(Sqrt, self).__init__(None, scope)

    def run(self, t):
        y = torch.sqrt(t)
        return y



# ------------------------------------ Polyadic -------------------------------------- #

class Concat(Craft):
    def __init__(self, scope='CONCAT'):
        super(Concat, self).__init__(None, scope)

    def run(self, t, axis=-1):
        y = torch.cat(t, dim=axis)
        return y



# ----------------------------------- Statistics ------------------------------------- #

class Clip(Craft):
    def __init__(self, intrinsic=False, scope='CLIP'):
        super(Clip, self).__init__(None, scope)
        self.intrinsic = intrinsic

    def run(self, x, ranges):
        if isinstance(ranges, tuple):
            assert len(ranges)==2
        else:
            ranges = (-ranges, ranges)
        if self.intrinsic:
            x.data = x.clamp(ranges[0], ranges[1])
            return
        else:
            return torch.clamp(x, ranges[0], ranges[1])



class Mean(Craft):
    def __init__(self, scope='MEAN'):
        super(Mean, self).__init__(None, scope)

    def run(self, x, axis=None):
        if axis is None:
            y = torch.mean(x)
        else:
            y = torch.mean(x, dim=axis)
        return y



class Max(Craft):
    def __init__(self, scope='MAX'):
        super(Max, self).__init__(None, scope)

    def run(self, x, axis=None):
        if axis is None:
            y = torch.max(x)
        else:
            y = torch.max(x, dim=axis)
        return y



class Min(Craft):
    def __init__(self, scope='MIN'):
        super(Min, self).__init__(None, scope)

    def run(self, x, axis=None):
        if axis is None:
            y = torch.min(x)
        else:
            y = torch.min(x, dim=axis)
        return y



class Sum(Craft):
    def __init__(self, scope='SUM'):
        super(Sum, self).__init__(None, scope)

    def run(self, x, axis=None):
        if axis is None:
            y = torch.sum(x)
        else:
            y = torch.sum(x, dim=axis)
        return y



class MatMul(Craft):
    def __init__(self, scope='MATMUL'):
        super(MatMul, self).__init__(None, scope)

    def run(self, x, y, in_batch=False):
        if in_batch:
            z = torch.bmm(x, y)
        else:
            z = torch.mm(x, y)
        return z



class Grad(Craft):
    def __init__(self, scope='GRAD'):
        super(Grad, self).__init__(None, scope)

    def run(self, i, o):
        grad_out_weight = torch.ones((i.shape[0], 1), device=i.device)
        g = torch.autograd.grad(o, i, grad_outputs=grad_out_weight,
                                retain_graph=True, create_graph=True, only_inputs=True)
        return g



# ------------------------------------- Resizer -------------------------------------- #

class Reshape(Craft):
    def __init__(self, scope='RESHAPE'):
        super(Reshape, self).__init__(None, scope)

    def run(self, x, shape):
        y = torch.reshape(x, shape)
        return y



class Permute(Craft):
    def __init__(self, scope='PERMUTE'):
        super(Permute, self).__init__(None, scope)

    def run(self, x, shape):
        y = x.permute(shape)
        return y



class Upscale(Craft):
    def __init__(self, scale: tuple, interp=NEAREST, scope='UPS'):
        super(Upscale, self).__init__(None, scope)
        self.fn = nn.Upsample(scale_factor=scale, mode=PT_INTERP[interp])

    def run(self, x):
        y = self.fn(x)
        return y



class MaxPool(Craft):
    def __init__(self, kernel: tuple, stride=2, padding=0, scope='MPOOL'):
        super(MaxPool, self).__init__(None, scope)
        dim = len(kernel)
        is_global = True if kernel[-1] < 0 else False
        if dim == 1:
            if is_global:
                pool_fn = nn.AdaptiveMaxPool1d
            else:
                pool_fn = nn.MaxPool1d
            pad_fn = nn.ConstantPad1d
        elif dim == 2:
            if is_global:
                pool_fn = nn.AdaptiveMaxPool2d
            else:
                pool_fn = nn.MaxPool2d
            pad_fn = nn.ConstantPad2d
        elif dim == 3:
            if is_global:
                pool_fn = nn.AdaptiveMaxPool3d
            else:
                pool_fn = nn.MaxPool3d
            pad_fn = nn.ConstantPad3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d pooling is not supported.' % dim)

        if isinstance(stride, int):
            stride = dim * [stride]

        self.pad = pad_fn(padding, 0)
        if is_global:
            assert padding == 0
            self.pool = pool_fn(tuple(dim * [1]))
        else:
            self.pool = pool_fn(kernel_size=kernel, stride=stride)

    def run(self, x):
        y = self.pool(self.pad(x))
        return y



class AvgPool(Craft):
    def __init__(self, kernel: tuple, stride=2, padding=0, scope='APOOL'):
        super(AvgPool, self).__init__(None, scope)
        dim = len(kernel)
        is_global = True if kernel[-1]<0 else False
        if dim == 1:
            if is_global:
                pool_fn = nn.AdaptiveAvgPool1d
            else:
                pool_fn = nn.AvgPool1d
            pad_fn = nn.ConstantPad1d
        elif dim == 2:
            if is_global:
                pool_fn = nn.AdaptiveAvgPool2d
            else:
                pool_fn = nn.AvgPool2d
            pad_fn = nn.ConstantPad2d
        elif dim == 3:
            if is_global:
                pool_fn = nn.AdaptiveAvgPool3d
            else:
                pool_fn = nn.AvgPool3d
            pad_fn = nn.ConstantPad3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d pooling is not supported.' % dim)

        if isinstance(stride, int):
            stride = dim * [stride]

        self.pad = pad_fn(padding, 0)
        if is_global:
            assert padding == 0
            self.pool = pool_fn(tuple(dim * [1]))
        else:
            self.pool = pool_fn(kernel_size=kernel, stride=stride)

    def run(self, x):
        y = self.pool(self.pad(x))
        return y



# ------------------------------------ Activation ------------------------------------ #

class Relu(Craft):
    def __init__(self, scope='RELU'):
        super(Relu, self).__init__(None, scope)
        self.actv = nn.ReLU()

    def run(self, x):
        y = self.actv(x)
        return y



class LRelu(Craft):
    def __init__(self, alpha=0.2, scope='LRELU'):
        super(LRelu, self).__init__(None, scope)
        self.actv = nn.LeakyReLU(alpha)

    def run(self, x):
        y = self.actv(x)
        return y



class Tanh(Craft):
    def __init__(self, scope='TANH'):
        super(Tanh, self).__init__(None, scope)
        self.actv = nn.Tanh()

    def run(self, x):
        y = self.actv(x)
        return y



class Sigm(Craft):
    def __init__(self, scope='SIGM'):
        super(Sigm, self).__init__(None, scope)
        self.actv = nn.Sigmoid()

    def run(self, x):
        y = self.actv(x)
        return y



class Sftm(Craft):
    def __init__(self, axis=-1, scope='SFTM'):
        super(Sftm, self).__init__(None, scope)
        self.actv = nn.Softmax(dim=axis)

    def run(self, x):
        y = self.actv(x)
        return y



class Sftp(Craft):
    def __init__(self, scope='SFTP'):
        super(Sftp, self).__init__(None, scope)
        self.actv = nn.Softplus(1)

    def run(self, x):
        y = self.actv(x)
        return y



# ------------------------------------ Distributing ------------------------------------ #

class Dropout(Craft):
    def __init__(self, p_drop, dim, scope='DROPOUT'):
        super(Dropout, self).__init__(None, scope)
        if dim == 1:
            dp_fn = nn.Dropout
        elif dim == 2:
            dp_fn = nn.Dropout2d
        elif dim == 3:
            dp_fn = nn.Dropout3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d BN is not supported.' % dim)
        self.dp = dp_fn(p=p_drop)

    def run(self, x):
        y = self.dp(x)
        return y



class BN(Craft):
    def __init__(self, out_chs, dim, mmnt=0.9, resilient=True, scope='BN'):
        super(BN, self).__init__(None, scope)
        if dim == 1:
            norm_fn = nn.BatchNorm1d
        elif dim == 2:
            norm_fn = nn.BatchNorm2d
        elif dim == 3:
            norm_fn = nn.BatchNorm3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d BN is not supported.' % dim)
        self.norm = norm_fn(out_chs, momentum=1 - mmnt, affine=resilient, eps=1e-5)

    def run(self, x):
        y = self.norm(x)
        return y



class CBN(Craft):
    def __init__(self, in_chs, out_chs, dim, mmnt=0.9, scope='CBN'):
        super(CBN, self).__init__(None, scope)
        if dim == 1:
            norm_fn = nn.BatchNorm1d
        elif dim == 2:
            norm_fn = nn.BatchNorm2d
        elif dim == 3:
            norm_fn = nn.BatchNorm3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d CN is not supported.' % dim)
        self.norm = norm_fn(out_chs, momentum=1 - mmnt, affine=False, eps=1e-5)
        self.relu = nn.ReLU()
        self.gamma_1 = nn.Linear(in_chs, in_chs // 2)
        self.gamma_2 = nn.Linear(in_chs // 2, out_chs)
        self.beta_1 = nn.Linear(in_chs, in_chs // 2)
        self.beta_2 = nn.Linear(in_chs // 2, out_chs)

    def run(self, x, z):
        y = self.norm(x)

        g = self.gamma_1(z)
        g = self.relu(g)
        g = self.gamma_2(g)

        b = self.beta_1(z)
        b = self.relu(b)
        b = self.beta_2(b)

        for _ in range(x.ndim - 2):
            g = g.unsqueeze(-1)
            b = b.unsqueeze(-1)

        self.weight = g
        self.bias = b
        y = self.weight * y + self.bias

        return y



class IN(Craft):
    def __init__(self, out_chs, dim, mmnt=0.9, resilient=True, scope='IN'):
        super(IN, self).__init__(None, scope)
        if dim == 1:
            norm_fn = nn.InstanceNorm1d
        elif dim == 2:
            norm_fn = nn.InstanceNorm2d
        elif dim == 3:
            norm_fn = nn.InstanceNorm3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d IN is not supported.' % dim)
        self.norm = norm_fn(out_chs, momentum=1 - mmnt, affine=resilient, eps=1e-5)

    def run(self, x):
        y = self.norm(x)
        return y



class CIN(Craft):
    def __init__(self, in_chs, out_chs, dim, mmnt=0.9, scope='CIN'):
        super(CIN, self).__init__(None, scope)
        if dim == 1:
            norm_fn = nn.InstanceNorm1d
        elif dim == 2:
            norm_fn = nn.InstanceNorm2d
        elif dim == 3:
            norm_fn = nn.InstanceNorm3d
        else:
            raise Exception('NEBULAE ERROR ⨷ %d-d CN is not supported.' % dim)
        self.norm = norm_fn(out_chs, momentum=1 - mmnt, affine=False, eps=1e-5)
        self.relu = nn.ReLU()
        self.gamma_1 = nn.Linear(in_chs, in_chs // 2)
        self.gamma_2 = nn.Linear(in_chs // 2, out_chs)
        self.beta_1 = nn.Linear(in_chs, in_chs // 2)
        self.beta_2 = nn.Linear(in_chs // 2, out_chs)

    def run(self, x, z):
        y = self.norm(x)

        g = self.gamma_1(z)
        g = self.relu(g)
        g = self.gamma_2(g)

        b = self.beta_1(z)
        b = self.relu(b)
        b = self.beta_2(b)

        for _ in range(x.ndim - 2):
            g = g.unsqueeze(-1)
            b = b.unsqueeze(-1)

        self.weight = g
        self.bias = b
        y = self.weight * y + self.bias

        return y



class LN(Craft):
    def __init__(self, norm_shape, resilient=True, scope='LN'):
        super(LN, self).__init__(None, scope)
        norm_shape = tuple([norm_shape[-1]] + [ns for ns in norm_shape[:-1]])
        self.norm = nn.LayerNorm(norm_shape, elementwise_affine=resilient, eps=1e-5)

    def run(self, x):
        y = self.norm(x)
        return y



class SN(Craft):
    def __init__(self, craft, niter=3, eps=1e-12, scope='SN'):
        super(SN, self).__init__(None, scope)
        self.name = 'weight'
        if isinstance(craft, (Conv, TransConv)):
            self.key = 'conv/'
        elif isinstance(craft, Dense):
            self.key = 'fc/'
        elif isinstance(craft, Embed):
            self.key = 'embd/'
        elif isinstance(craft, (BN, IN, LN)):
            self.key = 'norm/'
        elif isinstance(craft, (CBN, CIN)):
            self.key = ''
        else:
            raise Exception('NEBULAE ERROR ⨷ SN does not support %s layer.' % type(craft))
        self.craft = craft
        self.hull = craft[self.key]
        self.niter = niter
        self.eps = eps
        if not self._made_params():
            self._make_params()

    def l2normalize(self, v):
        return v / (v.norm() + self.eps)

    def _update_u_v(self):
        if not self._made_params():
            self._make_params()
        w = getattr(self.hull, self.name)
        u = getattr(self.hull, self.name + "_u")

        height = w.data.shape[0]
        for _ in range(self.niter):
            v = self.l2normalize(torch.mv(torch.t(w.view(height, -1).data), u))
            u = self.l2normalize(torch.mv(w.view(height, -1).data, v))

        setattr(self.hull, self.name + "_u", u)
        w.data = w.data / torch.dot(u, torch.mv(w.view(height, -1).data, v))

    def _made_params(self):
        return hasattr(self.hull, self.name + "_u")

    def _make_params(self):
        w = getattr(self.hull, self.name)

        height = w.data.shape[0]
        width = w.view(height, -1).data.shape[1]

        u = self.l2normalize(w.data.new(height).normal_(0, 1))

        self.hull.register_buffer(self.name + "_u", u)

    def run(self, *args, **kwargs):
        self._update_u_v()
        return self.craft.run(*args, **kwargs)



# ------------------------------------ Loss ------------------------------------ #

class MAE(Craft):
    def __init__(self, scope='MAE'):
        super(MAE, self).__init__(None, scope)
        self.cost = nn.L1Loss()

    def run(self, x, y):
        z = self.cost(x, y)
        return z



class MSE(Craft):
    def __init__(self, scope='MAE'):
        super(MSE, self).__init__(None, scope)
        self.cost = nn.MSELoss()

    def run(self, x, y):
        z = self.cost(x, y)
        return z



class SigmXE(Craft):
    def __init__(self, scope='SFTMXE'):
        super(SigmXE, self).__init__(None, scope)
        self.cost = nn.BCEWithLogitsLoss()

    def run(self, x, y):
        z = self.cost(x, y)
        return z



class SftmXE(Craft):
    def __init__(self, is_one_hot, scope='SFTMXE'):
        super(SftmXE, self).__init__(None, scope)
        self.ioh = is_one_hot
        self.cost = nn.CrossEntropyLoss()

    def run(self, x, y):
        if self.ioh:
            y = torch.argmax(y, dim=-1)
        z = self.cost(x, y)
        return z



# ------------------------------------ Metric ------------------------------------ #

class AccCls(Craft):
    def __init__(self, multi_class, is_one_hot, scope='ACCCLS'):
        super(AccCls, self).__init__(None, scope)
        if multi_class:
            assert not is_one_hot
        self.mulcls = multi_class
        self.ioh = is_one_hot

    def run(self, x, y):
        if self.mulcls: # include binary classification as well
            x = torch.round(x)
            correct = torch.mean((x == y).float(), dim=-1)
            z = torch.mean((correct == 1).float())
        else:
            if self.ioh:
                y = torch.argmax(y, dim=-1)
            x = torch.argmax(x, dim=-1)
            z = torch.mean((x == y).float())
        return z


# ------------------------------------ Optimizer ------------------------------------ #

class StepLR(object):
    def __init__(self, period, factor):
        self.period = period
        self.factor = factor

    def __call__(self, optimizer):
        return torch.optim.lr_scheduler.StepLR(optimizer, self.period, gamma=self.factor)



class PolyLR(object):
    def __init__(self, cutoff, power):
        self.cutoff = cutoff
        self.power = power

    def __call__(self, optimizer):
        lr = optimizer.defaults['lr']
        lr_update = lambda mile: ((lr - 1e-4) * (1 - min(mile / self.cutoff, 1)) ** self.power) + 1e-4
        return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_update)



class CosLR(object):
    def __init__(self, period):
        self.period = period

    def __call__(self, optimizer):
        return torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=self.period,
                                                          eta_min=optimizer.defaults['lr']*0.001)



class ExpLR(object):
    def __init__(self, period, factor):
        self.period = period
        self.factor = factor

    def __call__(self, optimizer):
        lr = optimizer.defaults['lr']
        lr_update = lambda mile: lr * self.factor ** (mile / self.period)
        return torch.optim.lr_scheduler.LambdaLR(optimizer, lr_update)



class WavyLR(object):
    def __init__(self, period, dampen):
        self.period = period
        self.dampen = dampen

    def __call__(self, optimizer):
        lr_update = lambda x: self.dampen ** (x - 1)
        return torch.optim.lr_scheduler.CyclicLR(optimizer, optimizer.defaults['lr'] / 2, optimizer.defaults['lr'],
                                                 step_size_up=self.period // 2, scale_fn=lr_update, scale_mode='cycle')



class Momentum(Craft):
    def __init__(self, hull, lr, mmnt=0.9, wd=0, lr_decay=None,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='MOMENTUM'):
        super(Momentum, self).__init__(None, scope)
        # select parameters await updating
        if update_scope is None:
            update_var = hull.parameters()
        else:
            if isinstance(update_scope, str):
                update_scope = [update_scope]
            update_var = []
            for us in update_scope:
                paths = us.split('/')
                craft = hull
                for p in paths:
                    craft = getattr(craft, p)
                update_var.append(craft.parameters())

        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit>0:
            nn.utils.clip_grad_value_(update_var, grad_limit)
        self.optz = torch.optim.SGD(update_var, lr=lr, momentum=mmnt, weight_decay=wd)
        if lr_decay is None:
            self.lr_decay = None
        else:
            self.lr_decay = lr_decay(self.optz)

    def run(self, target):
        self.optz.zero_grad()
        target.backward()
        self.cnt += 1
        if self.cnt == self.grad_accum:
            self.cnt = 0
            self.optz.step()
            if self.lr_decay is not None:
                self.lr_decay.step()



class Nesterov(Craft):
    def __init__(self, hull, lr, mmnt=0.9, wd=0, lr_decay=None,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='NESTEROV'):
        super(Nesterov, self).__init__(None, scope)
        # select parameters await updating
        if update_scope is None:
            update_var = hull.parameters()
        else:
            if isinstance(update_scope, str):
                update_scope = [update_scope]
            update_var = []
            for us in update_scope:
                paths = us.split('/')
                craft = hull
                for p in paths:
                    craft = getattr(craft, p)
                update_var.append(craft.parameters())

        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit > 0:
            nn.utils.clip_grad_value_(update_var, grad_limit)
        self.optz = torch.optim.SGD(update_var, lr=lr, momentum=mmnt, weight_decay=wd, nesterov=True)
        if lr_decay is None:
            self.lr_decay = None
        else:
            self.lr_decay = lr_decay(self.optz)

    def run(self, target):
        self.optz.zero_grad()
        target.backward()
        self.cnt += 1
        if self.cnt == self.grad_accum:
            self.cnt = 0
            self.optz.step()
            if self.lr_decay is not None:
                self.lr_decay.step()



class RMSProp(Craft):
    def __init__(self, hull, lr, mmnt=0., wd=0, lr_decay=None,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='RMSPROP'):
        super(RMSProp, self).__init__(None, scope)
        # select parameters await updating
        if update_scope is None:
            update_var = hull.parameters()
        else:
            if isinstance(update_scope, str):
                update_scope = [update_scope]
            update_var = []
            for us in update_scope:
                paths = us.split('/')
                craft = hull
                for p in paths:
                    craft = getattr(craft, p)
                update_var.append(craft.parameters())

        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit > 0:
            nn.utils.clip_grad_value_(update_var, grad_limit)
        self.optz = torch.optim.RMSprop(update_var, lr=lr, momentum=mmnt, weight_decay=wd)
        if lr_decay is None:
            self.lr_decay = None
        else:
            self.lr_decay = lr_decay(self.optz)

    def run(self, target):
        self.optz.zero_grad()
        target.backward()
        self.cnt += 1
        if self.cnt == self.grad_accum:
            self.cnt = 0
            self.optz.step()
            if self.lr_decay is not None:
                self.lr_decay.step()



class Adam(Craft):
    def __init__(self, hull, lr, mmnt1=0.9, mmnt2=0.999, wd=0, lr_decay=None,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='ADAM'):
        super(Adam, self).__init__(None, scope)
        # select parameters await updating
        if update_scope is None:
            update_var = hull.parameters()
        else:
            if isinstance(update_scope, str):
                update_scope = [update_scope]
            update_var = []
            for us in update_scope:
                paths = us.split('/')
                craft = hull
                for p in paths:
                    craft = getattr(craft, p)
                update_var.append(craft.parameters())

        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit > 0:
            nn.utils.clip_grad_value_(update_var, grad_limit)
        self.optz = torch.optim.Adam(update_var, lr=lr, betas=(mmnt1, mmnt2), weight_decay=wd)
        if lr_decay is None:
            self.lr_decay = None
        else:
            self.lr_decay = lr_decay(self.optz)

    def run(self, target):
        self.optz.zero_grad()
        target.backward()
        self.cnt += 1
        if self.cnt == self.grad_accum:
            self.cnt = 0
            self.optz.step()
            if self.lr_decay is not None:
                self.lr_decay.step()



class _AdaBelief(torch.optim.Optimizer):
    r"""Implements Adam algorithm.

    It has been proposed in `Adam: A Method for Stochastic Optimization`_.

    Arguments:
        params (iterable): iterable of parameters to optimize or dicts defining
            parameter groups
        lr (float, optional): learning rate (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of gradient and its square (default: (0.9, 0.999))
        eps (float, optional): term added to the denominator to improve
            numerical stability (default: 1e-8)
        weight_decay (float, optional): weight decay (L2 penalty) (default: 0)
        amsgrad (boolean, optional): whether to use the AMSGrad variant of this
            algorithm from the paper `On the Convergence of Adam and Beyond`_
            (default: False)

    .. _Adam\: A Method for Stochastic Optimization:
        https://arxiv.org/abs/1412.6980
    .. _On the Convergence of Adam and Beyond:
        https://openreview.net/forum?id=ryQu7f-RZ
    """

    def __init__(self, params, lr=1e-3, betas=(0.9, 0.999), eps=1e-8,
                 weight_decay=0, amsgrad=False):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        defaults = dict(lr=lr, betas=betas, eps=eps,
                        weight_decay=weight_decay, amsgrad=amsgrad)
        super(_AdaBelief, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(_AdaBelief, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('amsgrad', False)

    @torch.no_grad()
    def step(self, closure=None):
        """Performs a single optimization step.

        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            for p in group['params']:
                if p.grad is None:
                    continue
                grad = p.grad
                if grad.is_sparse:
                    raise RuntimeError(
                        'AdaBelief does not support sparse gradients, please consider SparseAdam instead')
                amsgrad = group['amsgrad']

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state['step'] = 0
                    # Exponential moving average of gradient values
                    state['exp_avg'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    # Exponential moving average of squared gradient values
                    state['exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format)
                    if amsgrad:
                        # Maintains max of all exp. moving avg. of sq. grad. values
                        state['max_exp_avg_sq'] = torch.zeros_like(p, memory_format=torch.preserve_format)

                exp_avg, exp_avg_sq = state['exp_avg'], state['exp_avg_sq']
                if amsgrad:
                    max_exp_avg_sq = state['max_exp_avg_sq']
                beta1, beta2 = group['betas']

                state['step'] += 1
                bias_correction1 = 1 - beta1 ** state['step']
                bias_correction2 = 1 - beta2 ** state['step']

                if group['weight_decay'] != 0:
                    grad = grad.add(p, alpha=group['weight_decay'])

                # Decay the first and second moment running average coefficient
                exp_avg.mul_(beta1).add_(grad, alpha=1 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad - exp_avg, grad - exp_avg, value=1 - beta2)
                if amsgrad:
                    # Maintains the maximum of all 2nd moment running avg. till now
                    torch.max(max_exp_avg_sq, exp_avg_sq, out=max_exp_avg_sq)
                    # Use the max. for normalizing running avg. of gradient
                    denom = (max_exp_avg_sq.sqrt() / sqrt(bias_correction2)).add_(group['eps'])
                else:
                    denom = (exp_avg_sq.sqrt() / sqrt(bias_correction2)).add_(group['eps'])

                step_size = group['lr'] / bias_correction1

                p.addcdiv_(exp_avg, denom, value=-step_size)

        return loss



class AdaBelief(Craft):
    def __init__(self, hull, lr, mmnt1=0.9, mmnt2=0.999, wd=0, lr_decay=None,
                 grad_limit=-1, grad_accum=1, update_scope=None, scope='ADABELIEF'):
        super(AdaBelief, self).__init__(None, scope)
        # select parameters await updating
        if update_scope is None:
            update_var = hull.parameters()
        else:
            if isinstance(update_scope, str):
                update_scope = [update_scope]
            update_var = []
            for us in update_scope:
                paths = us.split('/')
                craft = hull
                for p in paths:
                    craft = getattr(craft, p)
                update_var.append(craft.parameters())

        self.grad_accum = grad_accum
        self.cnt = 0
        if grad_limit > 0:
            nn.utils.clip_grad_value_(update_var, grad_limit)
        self.optz = _AdaBelief(update_var, lr=lr, betas=(mmnt1, mmnt2), weight_decay=wd)
        if lr_decay is None:
            self.lr_decay = None
        else:
            self.lr_decay = lr_decay(self.optz)

    def run(self, target):
        self.optz.zero_grad()
        target.backward()
        self.cnt += 1
        if self.cnt == self.grad_accum:
            self.cnt = 0
            self.optz.step()
            if self.lr_decay is not None:
                self.lr_decay.step()


#!/usr/bin/env python
'''
__init__
Created by Seria at 23/11/2018 10:30 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from .Classical.vgg import VGG_16
from .Classical.resnet import Resnet_V2_50, Resnet_V2_101, Resnet_V2_152
# from .senet import SE_RESNET_V2_50, SE_RESNET_V2_101, SE_RESNET_V2_152
from .ImgGen.gan import GAN
from .ImgGen.dcgan import DCGAN
from .ImgGen.fcgan import FCGAN
from .ImgGen.resgan import ResGAN
from .ImgGen.biggan import BigGAN
from .ImgGen.wgan import WGAN
from .ImgGen.wgan_div import WGANDiv
from .ImgGen.infogan import InfoGAN

BN = 20
CBN = 21
IN = 22
CIN = 23
LN = 24


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:38 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
import math
import numpy as np
import random as rand
from tensorflow import image as tfimg
import tensorflow as tf
from ..toolkit import byte2arr
import scipy.ndimage as ndimage


__all__ = ('Comburant', 'Random',
           'NEAREST', 'LINEAR', 'CUBIC', 'HORIZONTAL', 'VERTICAL',
           'Resize', 'Crop', 'Flip', 'Rotate',
           'Brighten', 'Contrast', 'Saturate', 'Hue')


NEAREST = 0
LINEAR = 1
CUBIC = 2

TF_INTERP = {NEAREST: 'nearest', LINEAR: 'bilinear', CUBIC: 'bicubic'}
SP_INTERP = {NEAREST: 0, LINEAR: 1, CUBIC: 3}

HORIZONTAL = 10
VERTICAL = 11




class Comburant(object):
    def __init__(self, *args, is_encoded=False):
        self.comburant = list(args)
        self.is_encoded = is_encoded

    def __call__(self, img):
        if self.is_encoded:
            # img = byte2arr(img)
            img = tfimg.decode_jpeg(img)
            img = tf.image.convert_image_dtype(img, tf.float32)
        for cbr in self.comburant:
            img = cbr(img)
        if not tf.is_tensor(img):
            img = tf.convert_to_tensor(img)
        return img



class Random(object):
    def __init__(self, p, comburant):
        self.p = p
        self.cbr = comburant

    def __call__(self, img):
        if rand.random() < self.p:
            return self.cbr(img)
        else:
            return img



class Resize(object):
    def __init__(self, size, interp=LINEAR):
        # size: (height, width)
        self.size = size
        self.interp = interp

    def __call__(self, img):
        return tfimg.resize(img, self.size, TF_INTERP[self.interp])



class Crop(object):
    def __init__(self, size: tuple, padding=(0, 0, 0, 0), area_ratio=(1, 1), aspect_ratio=(1, 1), interp=LINEAR):
        # size: (height, width)
        # padding: (left, top, right, bottom)
        self.size = size
        self.padding = padding
        self.area_ratio = area_ratio
        self.aspect_ratio = aspect_ratio
        self.interp = interp

    def __call__(self, img):
        h, w, c = img.shape
        l, t, r, b = self.padding
        _h = max(h + t + b, self.size[0])
        _w = max(w+l+r, self.size[1])
        img = tfimg.pad_to_bounding_box(img, t, l, _h, _w)
        h = _h
        w = _w

        if self.area_ratio == self.aspect_ratio == (1, 1):
            x = int(rand.random() * (w - self.size[1] + 1))
            y = int(rand.random() * (h - self.size[0] + 1))
            img = img[y:y+self.size[0], x:x+self.size[1]]
        else:
            sqrt_aspect_ratio = math.sqrt(rand.random()
                                          * (self.aspect_ratio[1] - self.aspect_ratio[0]) + self.aspect_ratio[0])
            sqrt_area_ratio = math.sqrt(rand.random()
                                        * (self.area_ratio[1] - self.area_ratio[0]) + self.area_ratio[0])
            cols = int(sqrt_area_ratio * w * sqrt_aspect_ratio)
            rows = int(sqrt_area_ratio * h / sqrt_aspect_ratio)
            x = int(rand.random() * (w - cols + 1))
            y = int(rand.random() * (h - rows + 1))
            img = tfimg.resize(img[y:y+rows, x:x+cols], self.size, TF_INTERP[self.interp])
        return img



class Flip(object):
    def __init__(self, axial):
        self.axial = axial

    def __call__(self, img):
        if self.axial == HORIZONTAL:
            img = tfimg.flip_up_down(img)
        elif self.axial == VERTICAL:
            img = tfimg.flip_left_right(img)
        else:
            raise Exception('NEBULAE ERROR ⨷ the invoked flip type is not defined or supported.')
        return img



class Rotate(object):
    def __init__(self, degree, intact=False, interp=NEAREST):
        '''
        Args
        intact: whether to keep image intact which might enlarge the output size
        '''
        self.degree = degree
        self.intact = intact
        self.interp = interp

    def call(self, img):
        degree = self.degree * (rand.random() * 2 - 1)
        return ndimage.rotate(img, degree, reshape=self.intact, order=SP_INTERP[self.interp])

    # def call(self, img):
    #     h, w, c = img.shape
    #     rot_center = (w / 2., h / 2.)
    #     degree = self.degree * (rand.random() * 2 - 1)
    #     angle = math.radians(degree)
    #     matrix = [
    #         round(math.cos(angle), 15), round(math.sin(angle), 15), 0.,
    #         round(-math.sin(angle), 15), round(math.cos(angle), 15), 0.,
    #     ]
    #
    #     def transform(x, y, matrix):
    #         (a, b, c, d, e, f) = matrix
    #         return a * x + b * y + c, d * x + e * y + f
    #
    #     matrix[2], matrix[5] = transform(-rot_center[0], -rot_center[1], matrix)
    #     matrix[2] += rot_center[0]
    #     matrix[5] += rot_center[1]
    #
    #     if self.intact:
    #         # calculate output size
    #         xx = []
    #         yy = []
    #         for x, y in ((0, 0), (w, 0), (w, h), (0, h)):
    #             x, y = transform(x, y, matrix)
    #             xx.append(x)
    #             yy.append(y)
    #         nw = math.ceil(max(xx)) - math.floor(min(xx))
    #         nh = math.ceil(max(yy)) - math.floor(min(yy))
    #
    #         # We multiply a translation matrix from the right.  Because of its
    #         # special form, this is the same as taking the image of the
    #         # translation vector as new translation vector.
    #         matrix[2], matrix[5] = transform(-(nw - w) / 2.0, -(nh - h) / 2.0, matrix)
    #     else:
    #         nw, nh = w, h
    #
    #     rot_img = np.zeros((nh, nw, c), dtype=img.dtype)
    #     zeros = np.zeros((c,), dtype=img.dtype)
    #     for i in range(nh):
    #         for j in range(nw):
    #             x, y = transform(j, i, matrix)
    #             x, y = round(x), round(y)
    #             if x >= w or y >= h or x < 0 or y < 0:
    #                 pix = zeros
    #             else:
    #                 pix = img[round(y), round(x)]
    #             rot_img[i, j] = pix
    #
    #     return rot_img

    def __call__(self, img):
        shape = img.shape
        img = tf.numpy_function(self.call, [img], tf.float32)
        img.set_shape(shape)
        return img



class Brighten(object):
    def __init__(self, range):
        self.range = range

    def __call__(self, img):
        return tfimg.random_brightness(img, self.range)



class Contrast(object):
    def __init__(self, range):
        self.range = range

    def __call__(self, img):
        return tfimg.random_contrast(img, max(0, 1-self.range), 1+self.range)



class Saturate(object):
    def __init__(self, range):
        self.range = range

    def __call__(self, img):
        return tfimg.random_saturation(img, max(0, 1-self.range), 1+self.range)



class Hue(object):
    def __init__(self, range):
        self.range = range

    def __call__(self, img):
        return tfimg.random_hue(img, self)


#!/usr/bin/env python
'''
Created by Seria at 02/11/2018 3:39 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from .generator import Generator
from ..toolkit.utility import _mergeFuel as merge
from ..toolkit.utility import _fillFuel as fill
from ..toolkit.utility import _deductFuel as deduct

__all__ = ('Generator', 'depot', 'merge', 'fill', 'deduct')


#!/usr/bin/env python
'''
layout_sheet
Created by Seria at 02/12/2018 1:20 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

from graphviz import Digraph
from ..astrobase.dock import Tensor
import os
import numpy as np

class BluePrint(object):

    def __init__(self, config=None, verbose=True, hidden=[]):
        if config is None:
            self.param = {'verbose': verbose, 'hidden': hidden}
        else:
            self.param = config
        self.layout = Digraph(comment='The Space Craft', format='jpg')
        self.layout.attr('node', shape='doublecircle')
        self.seen = []
        self.displayed = [] # already printed on screen
        self.shapes = {}
        self._await = False
        self._path = './_'

    def paint(self, archit, *slot):
        dummy = []
        core = os.environ.get('NEB_CORE', 'PYTORCH')
        for k,s,t in slot:
            if core == 'TENSORFLOW':
                import tensorflow as tf
                if isinstance(s, tuple):
                    data = np.random.rand(1, *s).astype(t)
                    dummy.append(Tensor(k, tf.covert_to_tensor(data)))
                else:
                    dummy.append(Tensor(k, s))
            # elif core == 'MXNET':
            #     from mxnet import nd
            #     if isinstance(s, tuple):
            #         data = np.random.rand(1, *s).astype(t)
            #         dummy.append(Tensor(k, nd.array(data)))
            #     else:
            #         dummy.append(Tensor(k, s))
            elif core == 'PYTORCH':
                import torch
                if isinstance(s, tuple):
                    data = np.random.rand(1, *s).astype(t)
                    dummy.append(Tensor(k, torch.from_numpy(data)))
                else:
                    dummy.append(Tensor(k, s))
            else:
                raise ValueError('NEBULAE ERROR ⨷ %s is an unsupported core.' % core)
        _ = archit(*tuple(dummy))
        self.archit = archit
        for pod in archit.pods:
            self._parse(pod)

    def _parse(self, pod):
        sym = pod.symbol
        comp = pod.comp
        if sym == '':
            return pod.name
        else:
            nodes = []
            for c in comp:
                node = self._parse(c)
                nodes.append(node)
                if node not in self.seen and node not in self.param['hidden']:
                    self.layout.node(node, node, shape='box')
                    self.seen.append(node)
                    if node not in self.archit.dict.keys():
                        self.shapes[node] = ' '
                    else:
                        self.shapes[node] = ' '+' x '.join([str(s) for s in self.archit[node].shape[1:]])
            if sym == '>':
                for i in range(len(nodes) - 1):
                    prev = nodes[i]
                    curr = nodes[i+1]
                    if prev.symbol=='^':
                        if curr.symbol=='^':
                            if prev not in self.seen:
                                self.layout.node(prev, '^', shape='circle')
                                self.seen.append(prev)
                            pivot = prev
                            for c in curr.comp:
                                self.layout.edge(pivot, c.name, label=self.shapes[pivot])
                        else:
                            assert curr.symbol==''
                            pivot = curr
                        for c in prev.comp:
                            self.layout.edge(c.name, pivot, label=self.shapes[c.name])
                    self.layout.edge(nodes[i], nodes[i + 1], label=self.shapes[nodes[i]])
                    if self.param['verbose']:
                        shape_out = self.shapes[nodes[i+1]]
                        print('| Assembling component\033[34m%32s\033[0m | Output%s |' % (nodes[i+1], shape_out))
                        print('+' + (63 + len(shape_out)) * '-' + '+')
                pod.name = nodes[-1]
                return nodes[-1]
            elif sym in ['+', '-', '*', '@', '&', '|', '^']:
                pivot = sym.join(nodes)
                pod.name = pivot
                if sym=='^':
                    return pivot
                else:
                    if pivot not in self.seen:
                        self.layout.node(pivot, sym, shape='circle')
                        self.seen.append(pivot)
                    for node in nodes:
                        self.layout.edge(node, pivot, label=self.shapes[node])
                    return pivot

    def log(self, log_dir=None, graph_name=None):
        core = os.environ.get('NEB_CORE', 'PYTORCH')
        # TODO: may need to be corrected for TF2.x
        if core=='TENSORFLOW' or self._await:
            if log_dir is None:
                log_dir = os.path.dirname(self._path)
            if graph_name is None:
                graph_name = os.path.basename(self._path)
            gv_name = os.path.join(log_dir, '%s.gv'%graph_name)
            self.layout.render(gv_name, view=False)
            os.rename(gv_name + '.jpg', os.path.join(log_dir, '%s.jpg'%graph_name))
            os.remove(gv_name)


#!/usr/bin/env python
'''
garage
Created by Seria at 03/01/2019 8:32 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock
from .architect import MLPG, MLPD



class WMLPD(dock.Craft):
    def __init__(self, in_shape, latent_dim, k, p, scope='WMLPD'):
        super(WMLPD, self).__init__(scope)
        self.k = k
        self.p = p
        self.backbone = MLPD(in_shape, latent_dim)
        self.cls = dock.Dense(latent_dim, 1)

        self.sum = dock.Sum()
        self.grad = dock.Grad()
        self.flat = dock.Reshape()

    def run(self, x, grad_penalty):
        self['input'] = x
        if grad_penalty:
            self['input'] = self.engine.coat(self['input'], as_const=False)

        self['out'] = self.cls(self.backbone(self['input']))

        if grad_penalty:
            # self.chain(self['input'])
            g = self.grad(self['input'], self['out'])[0]
            g = self.flat(g, (g.shape[0], -1))
            g = self.sum(g**2, -1)
            self['grad_norm'] = self.k * g ** (self.p/2)

            return self['out'], self['grad_norm']
        else:
            return self['out']



class WGANDiv(dock.Craft):
    def __init__(self, in_shape, latent_dim=128, k=2, p=6, scope='WGANDIV'):
        super(WGANDiv, self).__init__(scope)
        self.G = MLPG(in_shape, latent_dim)
        self.D = WMLPD(in_shape, latent_dim, k, p)

    def run(self):
        pass


#!/usr/bin/env python
'''
garage
Created by Seria at 03/01/2019 8:32 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock
from .architect import MLPG, MLPD


class WMLPD(dock.Craft):
    def __init__(self, in_shape, latent_dim, scope='WMLPD'):
        super(WMLPD, self).__init__(scope)
        self.backbone = MLPD(in_shape, latent_dim)
        self.cls = dock.Dense(latent_dim, 1)
        self.clipper = dock.Clip(intrinsic=True)

    def run(self, x):
        self['input'] = x
        self['out'] = self.cls(self.backbone(self['input']))

        return self['out']

    def clip(self, ranges):
        for p in self.params():
            _ = self.clipper(p, ranges)


class WGAN(dock.Craft):
    def __init__(self, in_shape, latent_dim=128, scope='WGAN'):
        super(WGAN, self).__init__(scope)
        self.G = MLPG(in_shape, latent_dim)
        self.D = WMLPD(in_shape, latent_dim)

    def run(self):
        pass


#!/usr/bin/env python
'''
setup
Created by Seria at 04/11/2018 10:50 AM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-

import setuptools

with open("Nebulae_Brochure.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nebulae",
    version="0.5.0",
    author="Seria",
    author_email="zzqsummerai@yeah.net",
    description="A novel and simple framework based on prevalent DL frameworks and other image processing libs."
                + " v0.5.0: the latest version takes the Big Two mainstream frameworks as backend,"
                + " and they share the same interfaces and arguments"
                + " i.e. Pytorch and Tensorflow. It is convenient for everyone to create nerworks and train,"
                + " no matter using single or multiple GPUs. Code blocks written in naive Pytorch or Tensorflow"
                + " are allowed to mix with Nebulae to work together.",
                # + " v0.4.20: fix a crucial typo which results in installation failure.",
                # + " v0.4.19: mute redundant INFO in multi-process.",
                # + " v0.4.18: add prep_fn argument in Tank.",
                # + " v0.4.17: remove mxnet core; add tf2 core which is under test.",
                # + " v0.4.16: engine is now a required argument for initializing Craft, so does as_const for Engine.coat;"
                # + " add depot.Random to control if the comburant event is gonna happen.",
                # + " v0.4.15: set 'grad_accum' argument in optimizer to control gradient accumulations.",
                # + " v0.4.14: add BigGAN and EMA wrapper;"
                # + " reform weight initializers and lr decayors as Class instead of flags.",
                # + " v0.4.13: reimplement Multiverse using built-in pytorch function; add SoftPlus layer.",
                # + " v0.4.12: optimize the interface of Res-GAN.",
                # + " v0.4.11: add Identity, MatMul, Permute, Embed layers and supports SN-Embed;"
                # + " orthogonal initialization is added.",
                # + " v0.4.10: add epoch, mile and MPE as new input arguments of INVIZ item in dashboard.",
                # + " v0.4.9: fix an error if input of GAN is a gray image.",
                # + " v0.4.8: add SN (Spectral Norm) and SN-GAN.",
                # + " v0.4.7: fix a bug of getting available GPUs.",
                # + " v0.4.6: add Instance Norm layer; slight changes on resnetGAN.",
                # + " v0.4.5: reorganize hangar.",
                # + " v0.4.4: add ResnetV2 and varieties of GANs in hangar;"
                # + " change input arguments in pooling layers for better use of global pooling.",
                # + " v0.4.3: add VGG16 in hangar.",
                # + " v0.4.2: optimize distributed training using Multiverse.",
                # + " v0.4.1: mute in all subprocesses.",
                # + " v0.4.0: a brand new version is released. it is more compatible with the backend framework"
                # + " so as to ease pain for transplanting code.",
                # + " v0.3.2: update implementations of data augmentation;"
                # + " users can build spacecraft as many as need and manage them in a spacedock.",
                # + " v0.3.1: fix an unexpected error when miles in DashBoard starts from a number greater than 1;"
                # + " fix a bug that strings cannot be stored using fuel generator;"
                # + " users are able to wrap their model written in core framework as a component by inheriting from OffTheShelf;"
                # + " garage is replaced by hangar and GAN is added to it;"
                # + " update spatial data augmentation methods;",
                # + " v0.3.0: fix a catastrophic bug that brings wrong shaped components in network;"
                # + " add new attribute, info, in Pod to serve another new tool named Imperative Symbol;"
                # + " now users can indicate to Dense layer which axis is to be projected;"
                # + " core modules of PyTorch is integrated now.",
                # + "v0.2.6: fix a bug that returns available gpu ids more than need;"
                # + "move the weight decay term to correct device;",
                # + "v0.2.5: unify convolution functions in different dimensions.",
                # + "v0.2.4: add 'rescale' option in FuelDepot.loadFuel().",
                # + "v0.2.3: fix a bug would cause error when decode hdf5 file.",
                # + "v0.2.2: remove the functionality of param:ckpt_scope with mxnet core.",
                # + "v0.2.1: mute re-initialization warning with mxnet core; add new way to save&load models.",
                # + "v0.2.0: now tensorflow and mxnet cores are completely supported."
                # + "we patch so much of it, and it is easier to take almost every module as a stand-alone plug-in.",
                # + " v0.1.21: set complete_last_batch as True when loading fuel to keep the last batch in same size as others.",
                # + " v0.1.20: optimize the execution of shuffling data.",
                # + " v0.1.19: if users want to print their results in a flexible way, they can pass a tailor function to Dash Board.",
                # + " v0.1.18: if_shuffle is modifiable for fuel tank now.",
                # + " v0.1.17: set MPE, volume and epoch as properties in fuel depot;"
                # + " if nothing is assigned to name while calling unloadFuel and modify of fuel depot, every tank will be manipulated",
                # + " v0.1.16: shorten some function names e.g. 'milesPerEpoch' to 'getMPE', 'editProperty' to 'modify'.",
                # + " v0.1.15: move merge, fill and deduct functions from toolkit to fuel.",
                # + " v0.1.14: fix a bug in toolkit.toDenseLabel.",
                # + " v0.1.13: new argument num_gpus indicates how many gpus you need.",
                # + " v0.1.12: change the quoting character of label file to |;"
                # + " add fillFuel and deductFuel in toolkit;"
                # + " users can store data with variable length;"
                # + " reorganize the logic for assembling components.",
                # + " v0.1.11: rename LayoutSheet as BluePrint;"
                # + " change the initial parts of Space Craft and Navigator;"
                # + " merge log function in Aerolog as an interface exposed by Navigator.",
                # + " v0.1.10: make it simpler to use Dash Board module alone.",
                # + " v0.1.9: be able to remove EXIF without modifying raw images.",
                # + " v0.1.7: users can generate hdf5 as several files since generating large dataset at once is risky;"
                # + " In addition, mergeFuel function is provided for merging multiple hdf5 files;"
                # + " users can remove EXIF in images while generating data file by setting keep_exif as False.",
                # + " v0.1.6: add SE Resnets to Garage;"
                # + " read image in RGB mode of which number of channel is 3.",
                # + " v0.1.5: fix a bug would return wrong device id when looking for available gpu.",
                # + " v0.1.1: change the way to register stage;"
                # + " fix a bug in DashBoard may draw points in wrong places.",
                # + " v0.1.0: A roughly complete version. New parts, Engine, Time Machine, Dashboard and Navigator are added.",
                # + " v0.0.18: fix a bug would repeatedly append variable scope.",
                # + " v0.0.17: network layout will be saved as image instead of pdf;"
                # + " add components: RESHAPE, SLICE, CLIP;"
                # + " implement Time Machine.",
                # + " v0.0.15: fix a bug would lead to wrong implementation of ** symbol."
                # + " v0.0.14: unable to assemble DUPLICATE component without name but passing existent name is allowed."
                # + " v0.0.13: allow users to assemble DUPLICATE component without passing name;"
                # + " add RESIZE component;"
                # + " shorten automatically generated component names."
                # + " v0.0.12: fix wrong implementation on data augmentation;"
                # + " FuelGenerator will keep original image size if width or height is not given.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/",
    packages=setuptools.find_packages(),
    install_requires=
    ['graphviz',
     'pynput',
     'h5py',
     'pillow',
     'piexif',
     'scipy',
     'tensorflow-probability',
     'numpy<1.19',
     'torch>=1.6.0'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


#!/usr/bin/env python
'''
architect
Created by Seria at 2020/11/18 1:42 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock

from math import ceil
import numpy as np


BN = 20
CBN = 21
IN = 22
CIN = 23
LN = 24



class MLPG(dock.Craft):
    def __init__(self, in_shape, latent_dim, engine, scope='MLPG'):
        super(MLPG, self).__init__(engine, scope)
        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C

        self.lrelu = dock.LRelu()
        self.fc_1 = dock.Dense(latent_dim, latent_dim)
        self.fc_2 = dock.Dense(latent_dim, latent_dim * 2)
        self.bn_2 = dock.BN(latent_dim * 2, dim=1, mmnt=0.8)
        self.fc_3 = dock.Dense(latent_dim * 2, latent_dim * 4)
        self.bn_3 = dock.BN(latent_dim * 4, dim=1, mmnt=0.8)
        self.fc_4 = dock.Dense(latent_dim * 4, latent_dim * 8)
        self.bn_4 = dock.BN(latent_dim * 8, dim=1, mmnt=0.8)

        self.fc_pixel = dock.Dense(latent_dim * 8, H * W * C)
        self.tanh = dock.Tanh()
        self.rect = dock.Reshape()

    def run(self, z):
        self['latent_code'] = z
        z = self.fc_1(self['latent_code'])
        z = self.lrelu(z)

        z = self.fc_2(z)
        z = self.bn_2(z)
        z = self.lrelu(z)

        z = self.fc_3(z)
        z = self.bn_3(z)
        z = self.lrelu(z)

        z = self.fc_4(z)
        z = self.bn_4(z)
        z = self.lrelu(z)

        z = self.fc_pixel(z)
        z = self.tanh(z)
        self['fake'] = self.rect(z, (-1, self.C, self.H, self.W))

        return self['fake']


class MLPD(dock.Craft):
    def __init__(self, in_shape, latent_dim, engine, scope='MLPD'):
        super(MLPD, self).__init__(engine, scope)
        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C

        self.lrelu = dock.LRelu()
        self.flat = dock.Reshape()
        self.fc_1 = dock.Dense(H * W * C, latent_dim * 2)
        self.fc_2 = dock.Dense(latent_dim * 2, latent_dim)

    def run(self, x):
        self['input'] = x
        x = self.flat(self['input'], (-1, self.C * self.H * self.W))
        x = self.fc_1(x)
        x = self.lrelu(x)
        x = self.fc_2(x)
        self['out'] = self.lrelu(x)

        return self['out']



class ConvG(dock.Craft):
    def __init__(self, in_shape, latent_dim, engine, scope='CONVG'):
        super(ConvG, self).__init__(engine, scope)
        H, W, C = in_shape
        assert H % 4 == 0 and W % 4 == 0
        self.H = H
        self.W = W
        self.C = C

        self.lrelu = dock.LRelu()
        self.up = dock.Upscale(2)

        self.fc_1 = dock.Dense(latent_dim, 128 * (H // 4) * (W // 4))
        self.bn_1 = dock.BN(128, dim=2)
        pad = dock.autoPad((H // 2, W // 2), (3, 3))
        self.conv_1 = dock.Conv(128, 128, (3, 3), padding=pad)
        self.bn_2 = dock.BN(128, dim=2, mmnt=0.8)
        pad = dock.autoPad((H, W), (3, 3))
        self.conv_2 = dock.Conv(128, 64, (3, 3), padding=pad)
        self.bn_3 = dock.BN(64, dim=2, mmnt=0.8)
        pad = dock.autoPad((H, W), (3, 3))
        self.conv_3 = dock.Conv(64, C, (3, 3), padding=pad)

        self.tanh = dock.Tanh()
        self.reshape = dock.Reshape()

    def run(self, z):
        self['latent_code'] = z
        z = self.fc_1(self['latent_code'])
        z = self.reshape(z, (-1, 128, self.H // 4, self.W // 4))

        z = self.bn_1(z)
        z = self.up(z)
        z = self.conv_1(z)

        z = self.bn_2(z)
        z = self.lrelu(z)
        z = self.up(z)
        z = self.conv_2(z)

        z = self.bn_3(z)
        z = self.lrelu(z)
        z = self.conv_3(z)

        self['fake'] = self.tanh(z)

        return self['fake']

class DeconvG(dock.Craft):
    def __init__(self, in_shape, latent_dim, engine, scope='DECONVG'):
        super(DeconvG, self).__init__(engine, cope)
        H, W, C = in_shape
        assert H%4==0 and W%4==0
        self.H = H
        self.W = W
        self.C = C

        self.lrelu = dock.LRelu()

        self.fc_1 = dock.Dense(latent_dim, 128 * (H // 4) * (W // 4))
        self.bn_1 = dock.BN(128, dim=2)
        pad = dock.autoPad((H // 2, W // 2), (3, 3), stride=2)
        self.conv_1 = dock.TransConv(128, 128, (H // 2, W // 2), kernel=(3, 3), stride=2, padding=pad)
        self.bn_2 = dock.BN(128, dim=2, mmnt=0.8)
        pad = dock.autoPad((H, W), (3, 3), stride=2)
        self.conv_2 = dock.TransConv(128, 64, (H, W), kernel=(3, 3), stride=2, padding=pad)
        self.bn_3 = dock.BN(64, dim=2, mmnt=0.8)
        pad = dock.autoPad((H, W), (3, 3))
        self.conv_3 = dock.Conv(64, C, (3, 3), padding=pad)

        self.tanh = dock.Tanh()
        self.reshape = dock.Reshape()

    def run(self, z):
        self['latent_code'] = z
        z = self.fc_1(self['latent_code'])
        z = self.reshape(z, (-1, 128, self.H//4, self.W//4))

        z = self.bn_1(z)
        z = self.conv_1(z)

        z = self.bn_2(z)
        z = self.lrelu(z)
        z = self.conv_2(z)

        z = self.bn_3(z)
        z = self.lrelu(z)
        z = self.conv_3(z)

        self['fake'] = self.tanh(z)

        return self['fake']

class ConvD(dock.Craft):
    def __init__(self, in_shape, engine, scope='CONVD'):
        super(ConvD, self).__init__(engine, scope)
        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C

        self.lrelu = dock.LRelu()
        self.flat = dock.Reshape()
        pad = dock.autoPad((H, W), (3, 3), stride=2)
        self.conv_1 = dock.Conv(C, 16, (3, 3), stride=2, padding=pad)
        pad = dock.autoPad((ceil(H / 2), ceil(W / 2)), (3, 3), stride=2)
        self.conv_2 = dock.Conv(16, 32, (3, 3), stride=2, padding=pad)
        self.bn_2 = dock.BN(32, dim=2, mmnt=0.8)
        pad = dock.autoPad((ceil(H / 4), ceil(W / 4)), (3, 3), stride=2)
        self.conv_3 = dock.Conv(32, 64, (3, 3), stride=2, padding=pad)
        self.bn_3 = dock.BN(64, dim=2, mmnt=0.8)
        pad = dock.autoPad((ceil(H / 8), ceil(W / 8)), (3, 3), stride=2)
        self.conv_4 = dock.Conv(64, 128, (3, 3), stride=2, padding=pad)
        self.bn_4 = dock.BN(128, dim=2, mmnt=0.8)

    def run(self, x):
        bs = x.shape[0]
        self['input'] = x
        x = self.conv_1(x)
        x = self.lrelu(x)

        x = self.conv_2(x)
        x = self.lrelu(x)
        x = self.bn_2(x)

        x = self.conv_3(x)
        x = self.lrelu(x)
        x = self.bn_3(x)

        x = self.conv_4(x)
        x = self.lrelu(x)
        x = self.bn_4(x)

        self['out'] = self.flat(x, (bs, -1))

        return self['out']



class ResBlock(dock.Craft):
    def __init__(self, in_shape, neck_chs, body_chs, stride=1, norm_fn=BN,
                 attention=False, spec_norm=False, w_init=dock.XavierNorm(), engine=None, scope='RESBLK'):
        super(ResBlock, self).__init__(engine, scope)
        H, W, C = in_shape
        self.norm_fn = norm_fn
        self.body_chs = body_chs
        self.cn = False
        if norm_fn == BN:
            self.n_1 = dock.BN(C, dim=2)
            self.n_2 = dock.BN(neck_chs, dim=2)
            # self.ds_n = dock.BN(self.body_chs, dim=2)
        elif norm_fn == IN:
            self.n_1 = dock.IN(C, dim=2)
            self.n_2 = dock.IN(neck_chs, dim=2)
            # self.ds_n = dock.IN(self.body_chs, dim=2)
        elif norm_fn == CBN:
            self.n_1 = dock.CBN(128, C, dim=2)
            self.n_2 = dock.CBN(128, neck_chs, dim=2)
            # self.ds_n = dock.CBN(128, self.body_chs, dim=2)
            self.cn = True
        elif norm_fn == CIN:
            self.n_1 = dock.CIN(128, C, dim=2)
            self.n_2 = dock.CIN(128, neck_chs, dim=2)
            # self.ds_n = dock.CIN(128, self.body_chs, dim=2)
            self.cn = True
        elif norm_fn == LN:
            self.n_1 = dock.LN(in_shape)
            self.n_2 = dock.LN((ceil(H / stride), ceil(W / stride), neck_chs))
            # self.ds_n = dock.LN((ceil(H / stride), ceil(W / stride), self.body_chs))
        elif norm_fn is None:
            self.n_1 = dock.Identity()
            self.n_2 = dock.Identity()
            # self.ds_n = dock.Identity()

        pad = dock.autoPad((H, W), (3, 3), stride=stride)
        if spec_norm:
            self.conv_1 = dock.SN(dock.Conv(C, neck_chs, (3, 3), stride=stride, padding=pad, w_init=w_init))
            self.conv_2 = dock.SN(dock.Conv(neck_chs, self.body_chs, (1, 1), w_init=w_init))
            self.ds_conv = dock.SN(dock.Conv(C, self.body_chs, (1, 1), stride, w_init=w_init))
        else:
            self.conv_1 = dock.Conv(C, neck_chs, (3, 3), stride=stride, padding=pad, w_init=w_init)
            self.conv_2 = dock.Conv(neck_chs, self.body_chs, (1, 1), w_init=w_init)
            self.ds_conv = dock.Conv(C, self.body_chs, (1, 1), stride, w_init=w_init)

        self.relu = dock.Relu()
        if attention and H==W==64:
            self.att = NonLocal((H // stride, W // stride, self.body_chs), self.body_chs//8, self.body_chs//2,
                                spec_norm, w_init)
        else:
            self.att = dock.Identity()

    def run(self, x):
        if self.cn:
            self['input'], self['noise'] = x
            y = self.n_1(self['input'], self['noise'])
            y = self.relu(y)
            y = self.conv_1(y)

            y = self.n_2(y, self['noise'])
            y = self.relu(y)
            y = self.conv_2(y)

            self['identity'] = self.ds_conv(self['input'])
        else:
            self['input'] = x
            y = self.n_1(self['input'])
            y = self.relu(y)
            y = self.conv_1(y)

            y = self.n_2(y)
            y = self.relu(y)
            y = self.conv_2(y)

            self['identity'] = self.ds_conv(self['input'])

        y += self['identity']
        y = self.att(y)

        return y


class ResG(dock.Craft):
    def __init__(self, in_shape, latent_dim, base_chs, norm_fn=BN, attention=False,
                 spec_norm=False, w_init=dock.XavierNorm(), engine=None, scope='RESG'):
        super(ResG, self).__init__(engine, scope)
        self.min_res = 4
        ichs = {128: [base_chs * c for c in (16, 16, 8, 4, 2)],
                256: [base_chs * c for c in (16, 16, 8, 8, 4, 2)]}
        ochs = {128: [base_chs * c for c in (16, 8, 4, 2, 1)],
                256: [base_chs * c for c in (16, 8, 8, 4, 2, 1)]}

        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C
        min_size = min(H, W)
        self.ichs = ichs[min_size]
        self.ochs = ochs[min_size]

        self.up = dock.Upscale(2)
        self.relu = dock.Relu()
        pad = dock.autoPad((H, W), (3, 3))
        if spec_norm:
            self.conv = dock.SN(dock.Conv(self.ochs[-1], C, (3, 3), padding=pad, w_init=w_init))
            self.fc = dock.SN(dock.Dense(latent_dim, self.ichs[0] * self.min_res * self.min_res, w_init=w_init))
        else:
            self.conv = dock.Conv(self.ochs[-1], C, (3, 3), padding=pad, w_init=w_init)
            self.fc = dock.Dense(latent_dim, self.ichs[0] * self.min_res * self.min_res, w_init=w_init)

        for i in range(len(self.ochs)):
            setattr(self, 'blk_%d'%i, ResBlock((self.min_res * 2**i, self.min_res * 2**i, self.ichs[i]),
                                                self.ichs[i]//2, self.ochs[i], norm_fn=norm_fn,
                                                attention=attention, spec_norm=spec_norm, w_init=w_init))

        self.cn = False
        if norm_fn == BN:
            self.nf = dock.BN(self.ochs[-1], dim=2)
        elif norm_fn == IN:
            self.nf = dock.IN(self.ochs[-1], dim=2)
        elif norm_fn == CBN:
            self.nf = dock.CBN(latent_dim, self.ochs[-1], dim=2)
            self.cn = True
        elif norm_fn == CIN:
            self.nf = dock.CIN(latent_dim, self.ochs[-1], dim=2)
            self.cn = True
        elif norm_fn == LN:
            self.nf = dock.LN((H, W, self.ochs[-1]))
        elif norm_fn is None:
            self.nf = dock.Identity()

        self.tanh = dock.Tanh()
        self.reshape = dock.Reshape()

    def run(self, z):
        self['latent_code'] = z
        z = self.fc(self['latent_code'])
        z = self.reshape(z, (-1, self.ichs[0], self.min_res, self.min_res))

        if self.cn:
            for i in range(len(self.ochs)):
                rb = getattr(self, 'blk_%d' % i)
                z = rb((z, self['latent_code']))
                n_shape = z.shape
                n_shape[1] = 1
                noise = self.engine.coat(np.random.normal(size=n_shape).astype('float32'))
                strength = self.engine.coat(np.array(0).astype('float32'), as_const=False)
                z += strength * noise
                z = self.up(z)

            z = self.nf(z, self['latent_code'])
        else:
            for i in range(len(self.ochs)):
                rb = getattr(self, 'blk_%d' % i)
                z = rb(z)
                z = self.up(z)

            z = self.nf(z)

        z = self.relu(z)
        z = self.conv(z)

        self['fake'] = self.tanh(z)

        return self['fake']

class ResD(dock.Craft):
    def __init__(self, in_shape, base_chs, norm_fn=BN, attention=False,
                 spec_norm=False, w_init=dock.XavierNorm(), engine=None, scope='RESD'):
        super(ResD, self).__init__(engine, scope)
        H, W, C = in_shape
        min_size = min(H, W)
        ichs = {128: [base_chs * c for c in (1, 2, 4, 8)],
                256: [base_chs * c for c in (1, 2, 4, 8, 8)]}
        ochs = {128: [base_chs * c for c in (2, 4, 8, 16)],
                256: [base_chs * c for c in (2, 4, 8, 8, 16)]}
        self.ichs = ichs[min_size]
        self.ochs = ochs[min_size]

        self.relu = dock.Relu()
        self.flat = dock.Reshape()
        pad = dock.autoPad((H, W), (3, 3))
        if spec_norm:
            self.conv = dock.SN(dock.Conv(C, self.ichs[0], (3, 3), padding=pad))
        else:
            self.conv = dock.Conv(C, self.ichs[0], (3, 3), padding=pad)

        if norm_fn == CBN:
            norm_fn = BN
        if norm_fn == CIN:
            norm_fn = IN

        for i in range(len(self.ichs)):
            pad = dock.autoPad((H // 2**i, W // 2**i), (2, 2), stride=2)
            setattr(self, 'pool_%d'%i, dock.AvgPool((2, 2), padding=pad))
            setattr(self, 'blk_%d'%i,
                    ResBlock((H // 2**(i+1), W // 2**(i+1), self.ichs[i]), self.ichs[i]//2, self.ochs[i],
                              norm_fn=norm_fn, attention=attention, spec_norm=spec_norm, w_init=w_init))

        if norm_fn == BN:
            self.nf = dock.BN(self.ochs[-1], dim=2)
        elif norm_fn == IN:
            self.nf = dock.IN(self.ochs[-1], dim=2)
        elif norm_fn == LN:
            self.nf = dock.LN((H // 2**len(self.ichs), W // 2**len(self.ichs), self.ochs[-1]))
        elif norm_fn is None:
            self.nf = dock.Identity()

    def run(self, x):
        bs = x.shape[0]
        self['input'] = x
        x = self.conv(x)
        x = self.relu(x)

        for i in range(len(self.ichs)):
            pl = getattr(self, 'pool_%d'%i)
            rb = getattr(self, 'blk_%d'%i)
            x = pl(x)
            x = rb(x)

        x = self.nf(x)
        x = self.relu(x)

        self['out'] = self.flat(x, (bs, -1))

        return self['out']




class NonLocal(dock.Craft):
    def __init__(self, in_shape, att_chs, neck_chs, spec_norm, w_init, engine, scope='NONLOCAL'):
        super(NonLocal, self).__init__(engine, scope)
        # Channel multiplier
        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C

        self.att_chs = att_chs
        self.neck_chs = neck_chs

        pad = dock.autoPad((H, W), (2, 2), 2)
        self.mpool_p = dock.MaxPool((2, 2), padding=pad)
        self.mpool_g = dock.MaxPool((2, 2), padding=pad)

        if spec_norm:
            self.theta = dock.SN(dock.Conv(self.C, self.att_chs, (1, 1), w_init=w_init))
            self.phi = dock.SN(dock.Conv(self.C, self.att_chs, (1, 1), w_init=w_init))
            self.g = dock.SN(dock.Conv(self.C, self.neck_chs, (1, 1), w_init=w_init))
            self.o = dock.SN(dock.Conv(self.neck_chs, self.C, (1, 1), w_init=w_init))
        else:
            self.theta = dock.Conv(self.C, self.att_chs, (1, 1), w_init=w_init)
            self.phi = dock.Conv(self.C, self.att_chs, (1, 1), w_init=w_init)
            self.g = dock.Conv(self.C, self.neck_chs, (1, 1), w_init=w_init)
            self.o = dock.Conv(self.neck_chs, self.C, (1, 1), w_init=w_init)

        self.flat = dock.Reshape()
        self.perm = dock.Permute()
        self.sftm = dock.Sftm()
        self.mm = dock.MatMul()

    def forward(self, x):
        # Apply convs
        theta = self.theta(x)
        phi = self.phi(x)
        phi = self.mpool_p(phi)
        g = self.g(x)
        g = self.mpool_g(g)
        # Perform reshapes
        theta = self.flat(theta, (-1, self.att_chs, self.H * self.W))
        phi = self.flat(phi, (-1, self.att_chs, self.H * self.W // 4))
        g = self.flat(g, (-1, self.neck_chs, self.H * self.W // 4))
        # Matmul and softmax to get attention maps
        theta = self.perm(theta, (0, 2, 1))
        alpha = self.mm(theta, phi, in_batch=True)
        alpha = self.sftm(alpha)
        # Attention map times g path
        alpha = self.perm(alpha, (0, 2, 1))
        o = self.mm(g, alpha, in_batch=True)
        o = self.flat(o, (-1, self.neck_chs, self.H, self.W))
        o = self.o(o)
        return o + x





class BigResBlock(dock.Craft):
    def __init__(self, in_shape, neck_chs, body_chs, stride=1, norm_fn=BN, ups=None, downs=None,
                 attention=False, spec_norm=False, w_init=dock.XavierNorm(), engine=None, scope='BIGRESBLK'):
        super(BigResBlock, self).__init__(engine, scope)
        H, W, C = in_shape
        self.norm_fn = norm_fn
        self.body_chs = body_chs
        assert ups is None or downs is None
        self.ups = ups
        self.downs = downs
        self.cat = dock.Concat()

        self.cn = False
        if norm_fn == BN:
            self.n_1 = dock.BN(C, dim=2)
            self.n_2 = dock.BN(neck_chs, dim=2)
            # self.ds_n = dock.BN(self.body_chs, dim=2)
        elif norm_fn == IN:
            self.n_1 = dock.IN(C, dim=2)
            self.n_2 = dock.IN(neck_chs, dim=2)
            # self.ds_n = dock.IN(self.body_chs, dim=2)
        elif norm_fn == CBN:
            self.n_1 = dock.CBN(128, C, dim=2)
            self.n_2 = dock.CBN(128, neck_chs, dim=2)
            # self.ds_n = dock.CBN(128, self.body_chs, dim=2)
            self.cn = True
        elif norm_fn == CIN:
            self.n_1 = dock.CIN(128, C, dim=2)
            self.n_2 = dock.CIN(128, neck_chs, dim=2)
            # self.ds_n = dock.CIN(128, self.body_chs, dim=2)
            self.cn = True
        elif norm_fn == LN:
            self.n_1 = dock.LN(in_shape)
            self.n_2 = dock.LN((ceil(H / stride), ceil(W / stride), neck_chs))
            # self.ds_n = dock.LN((ceil(H / stride), ceil(W / stride), self.body_chs))
        elif norm_fn is None:
            self.n_1 = dock.Identity()
            self.n_2 = dock.Identity()
            # self.ds_n = dock.Identity()

        pad = dock.autoPad((H, W), (3, 3), stride=stride)
        if spec_norm:
            self.conv_1 = dock.SN(dock.Conv(C, neck_chs, (3, 3), stride=stride, padding=pad, w_init=w_init))
            self.conv_2 = dock.SN(dock.Conv(neck_chs, self.body_chs, (1, 1), w_init=w_init))
            self.ds_conv = dock.SN(dock.Conv(C, self.body_chs, (1, 1), stride, w_init=w_init))
        else:
            self.conv_1 = dock.Conv(C, neck_chs, (3, 3), stride=stride, padding=pad, w_init=w_init)
            self.conv_2 = dock.Conv(neck_chs, self.body_chs, (1, 1), w_init=w_init)
            if downs is None:
                self.rs_conv = dock.Conv(C, self.body_chs, (1, 1), stride, w_init=w_init)
            elif ups is None:
                self.rs_conv = dock.Conv(C, self.body_chs - C, (1, 1), stride, w_init=w_init)

        self.relu = dock.Relu()
        if attention and H == W == 64:
            self.att = NonLocal((H // stride, W // stride, self.body_chs), self.body_chs // 8, self.body_chs // 2,
                                spec_norm, w_init)
        else:
            self.att = dock.Identity()

    def shortcut(self, x):
        if self.downs is None:  # upscale
            return self.ups(self.rs_conv(x))
        elif self.ups is None:  # pooling
            x = self.downs(x)
            h = self.rs_conv(x)
            return self.cat((x, h), axis=1)

    def run(self, x):
        if self.cn:
            self['input'], self['noise'] = x
            y = self.n_1(self['input'], self['noise'])
            y = self.relu(y)
            if self.downs is None:  # upscale
                y = self.ups(y)
            y = self.conv_1(y)

            y = self.n_2(y, self['noise'])
            y = self.relu(y)
            if self.ups is None:  # pooling
                y = self.downs(y)
            y = self.conv_2(y)

            self['identity'] = self.shortcut(self['input'])
        else:
            self['input'] = x
            y = self.n_1(self['input'])
            y = self.relu(y)
            if self.downs is None:  # upscale
                y = self.ups(y)
            y = self.conv_1(y)

            y = self.n_2(y)
            y = self.relu(y)
            if self.ups is None:  # pooling
                y = self.downs(y)
            y = self.conv_2(y)

            self['identity'] = self.shortcut(self['input'])

        y += self['identity']
        y = self.att(y)

        return y


class BigG(dock.Craft):
    def __init__(self, in_shape, latent_dim, base_chs, norm_fn=BN, attention=False,
                 spec_norm=False, w_init=dock.XavierNorm(), engine=None, scope='BIGG'):
        super(BigG, self).__init__(engine, scope)
        self.min_res = 8
        ichs = {128: [base_chs * c for c in (16, 8, 4, 2)]}
        ochs = {128: [base_chs * c for c in (8, 4, 2, 1)]}

        H, W, C = in_shape
        self.H = H
        self.W = W
        self.C = C
        min_size = min(H, W)
        self.ichs = ichs[min_size]
        self.ochs = ochs[min_size]

        up = dock.Upscale(2)
        self.relu = dock.Relu()
        pad = dock.autoPad((H, W), (3, 3))
        if spec_norm:
            self.conv = dock.SN(dock.Conv(self.ochs[-1], C, (3, 3), padding=pad, w_init=w_init))
            self.fc = dock.SN(dock.Dense(latent_dim, self.ichs[0] * self.min_res * self.min_res, w_init=w_init))
        else:
            self.conv = dock.Conv(self.ochs[-1], C, (3, 3), padding=pad, w_init=w_init)
            self.fc = dock.Dense(latent_dim, self.ichs[0] * self.min_res * self.min_res, w_init=w_init)

        for i in range(len(self.ochs)):
            setattr(self, 'blk_%d' % i, BigResBlock((self.min_res * 2 ** i, self.min_res * 2 ** i, self.ichs[i]),
                                                     self.ichs[i] // 2, self.ochs[i], norm_fn=norm_fn, ups=up,
                                                     attention = attention, spec_norm = spec_norm, w_init = w_init))

            self.cn = False
            if norm_fn == BN:
                self.nf = dock.BN(self.ochs[-1], dim=2)
            elif norm_fn == IN:
                self.nf = dock.IN(self.ochs[-1], dim=2)
            elif norm_fn == CBN:
                self.nf = dock.CBN(latent_dim, self.ochs[-1], dim=2)
                self.cn = True
            elif norm_fn == CIN:
                self.nf = dock.CIN(latent_dim, self.ochs[-1], dim=2)
                self.cn = True
            elif norm_fn == LN:
                self.nf = dock.LN((H, W, self.ochs[-1]))
            elif norm_fn is None:
                self.nf = dock.Identity()

            self.tanh = dock.Tanh()
            self.reshape = dock.Reshape()

    def run(self, z):
        self['latent_code'] = z
        z = self.fc(self['latent_code'])
        z = self.reshape(z, (-1, self.ichs[0], self.min_res, self.min_res))

        if self.cn:
            for i in range(len(self.ochs)):
                rb = getattr(self, 'blk_%d' % i)
                z = rb((z, self['latent_code']))

            z = self.nf(z, self['latent_code'])
        else:
            for i in range(len(self.ochs)):
                rb = getattr(self, 'blk_%d' % i)
                z = rb(z)

            z = self.nf(z)

        z = self.relu(z)
        z = self.conv(z)

        self['fake'] = self.tanh(z)

        return self['fake']

class BigD(dock.Craft):
    def __init__(self, in_shape, base_chs, norm_fn=BN, attention=False,
                 spec_norm=False, w_init=dock.XavierNorm(), engine=None, scope='BIGD'):
        super(BigD, self).__init__(engine, scope)
        H, W, C = in_shape
        min_size = min(H, W)
        ichs = {128: [base_chs * c for c in (1, 2, 4, 8)]}
        ochs = {128: [base_chs * c for c in (2, 4, 8, 16)]}
        self.ichs = ichs[min_size]
        self.ochs = ochs[min_size]

        self.relu = dock.Relu()
        self.flat = dock.Reshape()
        pad = dock.autoPad((H, W), (3, 3))
        if spec_norm:
            self.conv = dock.SN(dock.Conv(C, self.ichs[0], (3, 3), padding=pad))
        else:
            self.conv = dock.Conv(C, self.ichs[0], (3, 3), padding=pad)

        if norm_fn == CBN:
            norm_fn = BN
        if norm_fn == CIN:
            norm_fn = IN

        for i in range(len(self.ichs)):
            pad = dock.autoPad((H // 2 ** i, W // 2 ** i), (2, 2), stride=2)
            #             setattr(self, 'pool_%d'%i, dock.AvgPool((2, 2), padding=pad))
            pool = dock.AvgPool((2, 2), padding=pad)
            setattr(self, 'blk_%d' % i,
                    BigResBlock((H // 2 ** (i + 1), W // 2 ** (i + 1), self.ichs[i]), self.ichs[i] // 2,
                                 self.ochs[i], norm_fn=norm_fn, downs=pool, attention=attention,
                                 spec_norm=spec_norm, w_init=w_init))

        if norm_fn == BN:
            self.nf = dock.BN(self.ochs[-1], dim=2)
        elif norm_fn == IN:
            self.nf = dock.IN(self.ochs[-1], dim=2)
        # elif norm_fn == CBN:
        #     self.nf = dock.CBN(128, self.ochs[-1], dim=2)
        #     self.cn = True
        # elif norm_fn == CIN:
        #     self.nf = dock.CIN(128, self.ochs[-1], dim=2)
        #     self.cn = True
        elif norm_fn == LN:
            self.nf = dock.LN((H // 2 ** len(self.ichs), W // 2 ** len(self.ichs), self.ochs[-1]))
        elif norm_fn is None:
            self.nf = dock.Identity()

    def run(self, x):
        bs = x.shape[0]
        self['input'] = x
        x = self.conv(x)
        x = self.relu(x)

        for i in range(len(self.ichs)):
            rb = getattr(self, 'blk_%d' % i)
            x = rb(x)

        x = self.nf(x)
        x = self.relu(x)

        self['out'] = self.flat(x, (bs, -1))

        return self['out']


#!/usr/bin/env python
'''
garage
Created by Seria at 03/01/2019 8:32 PM
Email: zzqsummerai@yeah.net

                    _ooOoo_
                  o888888888o
                 o88`_ . _`88o
                 (|  0   0  |)
                 O \   。   / O
              _____/`-----‘\_____
            .’   \||  _ _  ||/   `.
            |  _ |||   |   ||| _  |
            |  |  \\       //  |  |
            |  |    \-----/    |  |
             \ .\ ___/- -\___ /. /
         ,--- /   ___\<|>/___   \ ---,
         | |:    \    \ /    /    :| |
         `\--\_    -. ___ .-    _/--/‘
   ===========  \__  NOBUG  __/  ===========
   
'''
# -*- coding:utf-8 -*-
from ... import dock

from math import ceil

class VGG_16(dock.Craft):
    def __init__(self, in_shape, p_drop=0.5, engine=None, scope='VGG_16'):
        super(VGG_16, self).__init__(engine, scope)
        self.p_drop = p_drop
        H, W, C = in_shape
        self.relu = dock.Relu()
        pad = dock.autoPad((H, W), (3, 3))
        self.conv_1_1 = dock.Conv(C, 8, (3, 3), padding=pad)
        self.conv_1_2 = dock.Conv(8, 64, (3, 3), padding=pad)
        pad = dock.autoPad((H, W), (2, 2), stride=2)
        self.mpool_1 = dock.MaxPool((2, 2), padding=pad)

        pad = dock.autoPad((ceil(H / 2), ceil(W / 2)), (3, 3))
        self.conv_2_1 = dock.Conv(64, 128, (3, 3), padding=pad)
        self.conv_2_2 = dock.Conv(128, 128, (3, 3), padding=pad)
        pad = dock.autoPad((ceil(H / 2), ceil(W / 2)), (2, 2), stride=2)
        self.mpool_2 = dock.MaxPool((2, 2), padding=pad)

        pad = dock.autoPad((ceil(H / 4), ceil(W / 4)), (3, 3))
        self.conv_3_1 = dock.Conv(128, 256, (3, 3), padding=pad)
        self.conv_3_2 = dock.Conv(256, 256, (3, 3), padding=pad)
        self.conv_3_3 = dock.Conv(256, 256, (3, 3), padding=pad)
        pad = dock.autoPad((ceil(H / 4), ceil(W / 4)), (2, 2), stride=2)
        self.mpool_3 = dock.MaxPool((2, 2), padding=pad)

        pad = dock.autoPad((ceil(H / 8), ceil(W / 8)), (3, 3))
        self.conv_4_1 = dock.Conv(256, 512, (3, 3), padding=pad)
        self.conv_4_2 = dock.Conv(512, 512, (3, 3), padding=pad)
        self.conv_4_3 = dock.Conv(512, 512, (3, 3), padding=pad)
        pad = dock.autoPad((ceil(H / 8), ceil(W / 8)), (2, 2), stride=2)
        self.mpool_4 = dock.MaxPool((2, 2), padding=pad)

        pad = dock.autoPad((ceil(H / 16), ceil(W / 16)), (3, 3))
        self.conv_5_1 = dock.Conv(512, 512, (3, 3), padding=pad)
        self.conv_5_2 = dock.Conv(512, 512, (3, 3), padding=pad)
        self.conv_5_3 = dock.Conv(512, 512, (3, 3), padding=pad)
        pad = dock.autoPad((ceil(H / 16), ceil(W / 16)), (2, 2), stride=2)
        self.mpool_5 = dock.MaxPool((2, 2), padding=pad)

        self.dropout = dock.Dropout(0.2, dim=2)
        self.glob_conv_1 = dock.Conv(512, 4096, (ceil(H / 32), ceil(W / 32)))
        self.glob_conv_2 = dock.Conv(4096, 4096, (1, 1))



    def run(self, x):
        self['input'] = x
        c1 = self.conv_1_1(self['input'])
        c1 = self.relu(c1)
        c1 = self.conv_1_2(c1)
        c1 = self.relu(c1)
        self['conv_1'] = self.mpool_1(c1)

        c2 = self.conv_2_1(self['conv_1'])
        c2 = self.relu(c2)
        c2 = self.conv_2_2(c2)
        c2 = self.relu(c2)
        self['conv_2'] = self.mpool_2(c2)

        c3 = self.conv_3_1(self['conv_2'])
        c3 = self.relu(c3)
        c3 = self.conv_3_2(c3)
        c3 = self.relu(c3)
        c3 = self.conv_3_3(c3)
        c3 = self.relu(c3)
        self['conv_3'] = self.mpool_3(c3)

        c4 = self.conv_4_1(self['conv_3'])
        c4 = self.relu(c4)
        c4 = self.conv_4_2(c4)
        c4 = self.relu(c4)
        c4 = self.conv_4_3(c4)
        c4 = self.relu(c4)
        self['conv_4'] = self.mpool_4(c4)

        c5 = self.conv_5_1(self['conv_4'])
        c5 = self.relu(c5)
        c5 = self.conv_5_2(c5)
        c5 = self.relu(c5)
        c5 = self.conv_5_3(c5)
        c5 = self.relu(c5)
        self['conv_5'] = self.mpool_5(c5)

        gc = self.glob_conv_1(self['conv_5'])
        gc = self.dropout(gc)
        gc = self.glob_conv_2(gc)
        self['gconv'] = self.dropout(gc)

        return self['gconv']